/* Copyright (c) 2014-2019, Mike Smith
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * o  Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * o  Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * o  Neither the name of the Mike Smith/N4VLF nor the names of its
 *    contributors may be used to endorse or promote products derived from
 *    this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <termios.h>
#include <sys/select.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/param.h>
#include <time.h>
#include <math.h>
#include <pwd.h>
#include <ctype.h>

/* 
   TODO:
   Make substitutions dynamic, before the run of the macro.
   Add substitution for current working directory
   Add sending a local file out serial port
*/

/* The following pragmas quiet warnings on Apple's version of GCC that are */
/* overly strict for this application. */
#ifdef __APPLE_CC__
#pragma GCC diagnostic ignored "-Wstrncat-size"
#pragma GCC diagnostic ignored "-Wpointer-sign"
#endif


/*----------------------------------------------------------------------------
 * COnstants & Definitions
 * --------------------------------------------------------------------------*/
/* Set the command key that activates command mode in the terminal. */
#define COMMAND_KEY       0x11     /* CTL-Q */
#define COMMAND_KEY_STR   "CTL-Q"
#define MACRO_FILE        ".binterm"

#define INBUF_SIZE        1025     /* command input buffer size for command */
                                   /* mode */

/* line ending format constants, used to set interpretation of end-of-line */
/* characters from serial input */
#define END_NL            1
#define END_CR            2
#define END_CRLF          3

/*
 * Baud strings and numeric constant values. Used in setting/reporting 
 * terminal parameters.
 */
#ifdef __APPLE_CC__
#define BAUD_VALUE_COUNT  18
#else
#define BAUD_VALUE_COUNT  21
#endif

char *bstr[] = { "50", "75", "110", "134", "150", "200", "300", "600",	\
		   "1200", "1800", "2400", "4800", "9600", "19200", "38400", \
		   "57600", "115200", "230400", "460800", "500000", "576000" };
#ifdef __APPLE_CC__
unsigned int bvals[] = { B50, B75, B110, B134, B150, B200, B300, B600,	\
			   B1200, B1800, B2400, B4800, B9600, B19200,	\
			   B38400, B57600, B115200, B230400 };
#else
unsigned int bvals[] = { B50, B75, B110, B134, B150, B200, B300, B600,	\
			   B1200, B1800, B2400, B4800, B9600, B19200,	\
			   B38400, B57600, B115200, B230400, B460800,	\
			   B500000, B576000 };
#endif


/*----------------------------------------------------------------------------
 * Function Prototypes
 * --------------------------------------------------------------------------*/
void parse_params(int, char **);
unsigned int convert_baud(char *);
int open_serial(void);
int open_console(void);
void terminal(void);
unsigned char parse_hex2dec(char *);
void set_defaults(void);
void dot_file(void);
void read_dot(char *);
void substitute_in_macros(void);


/*----------------------------------------------------------------------------
 * Global variables
 * --------------------------------------------------------------------------*/
unsigned int baudrate, databits, parity, stopbits, serial_fd, console_fd;
unsigned int log_fd, line_end, hex_mode, hw_flow, hex_count, logging_mode;
char *device;
char *macros[10];
struct termios ttysave; /* for restoring console when we exit */
struct termios ttystate; /* for storing 'raw' mode - see open_console() */


int main(int argc, char *argv[])
{
  /* set up initial defaults */
  set_defaults();

  /* see if there is a macro/defaults settings file */
  dot_file();

  /* handle command line arguments */
  parse_params(argc,argv);
  
  /* try to open serial device */
  if( !open_serial() )
    {
      fprintf(stderr,"Unable to open serial device '%s'\n",device);
      exit(-1);
    }

  /* open the 'screen' (console) */
  if( !open_console() )
    {
      close(serial_fd);
      fprintf(stderr,"Unable to set up terminal.\n");
      exit(-1);
    }

  /* perform substitutions on macros - baud rate and device will not change */
  /* during run */
  substitute_in_macros();

  /* go into terminal mode loop */
  terminal();

  fprintf(stdout,"Closing serial device %s\r\n",device);
  close(serial_fd);

  /* close log if it's open */
  if( logging_mode && (log_fd != -1) )
    {
      fprintf(stdout,"Closing log file.\r\n");
      close(log_fd);
    }
  
  /* restore terminal state */
  fprintf(stdout,"Restoring terminal state.\r\n");
  tcsetattr(console_fd, TCSANOW, &ttysave);
  close(console_fd);

  return 0;
}


/*----------------------------------------------------------------------------
 * FUNCTION: parse_params
 * PARAMS  : Count of command line arguments, point to string array
 *           of command line arguments
 * RETURNS : None
 * CALLS   : fprintf(), toupper(), convert_baud(), strdup()
 * DESCR   : Parse command line arguments if given. Otherwise, print
 *           a help message showing available options.
 * --------------------------------------------------------------------------*/
void parse_params(int argc, char **argv)
{
  unsigned int i;
  unsigned char *p;

  if((argc<2) && (device==NULL))
    {
      fprintf(stderr,"Usage: %s -c device [options]\n\n",argv[0]);
      fprintf(stderr,"-b baudrate [9600 default]\n");
      fprintf(stderr,"-d 5,6,7 or 8 data bits [8 default]\n");
      fprintf(stderr,"-p N,E or O parity [N default]\n");
      fprintf(stderr,"-s 1 or 2 stop bits [1 default]\n");
      fprintf(stderr,"-c device\n");
      fprintf(stderr,"-l N(ewline), R(eturn), or C(RLF) line ending "
	      "[N default]\n");
      fprintf(stderr,"-h enable HW flow control\n");
      exit(0);
    }
    
    for(i=1;i<argc;i++)
      {
	if(argv[i][0] == '-')
	  {
	    switch(toupper(argv[i][1]))
	      {
	      case 'L': /* line ending */
		i++;
		switch(toupper(argv[i][0]))
		  {
		  case 'N': line_end=END_NL;
		    break;
		  case 'R': line_end=END_CR;
		    break;
		  case 'C': line_end=END_CRLF;
		    break;
		  }
		break;
	      case 'B': /* set baudrate. */
		i++;
		if( !(baudrate=convert_baud(argv[i])) )
		  {
		    fprintf(stderr,"Unable to parse baudrate '%s'\n",argv[i]);
		    exit(-1);
		  }
		break;
	      case 'D': /* data bits */
		i++;
		if(argv[i][0]=='5') databits=CS5;
		if(argv[i][0]=='6') databits=CS6;
		if(argv[i][0]=='7') databits=CS7;
		if(argv[i][0]=='8') databits=CS8;
		break;
	      case 'S': /* stop bits */
		i++;
		if(argv[i][0]=='1') stopbits=~CSTOPB;
		if(argv[i][0]=='2') stopbits=CSTOPB;
		break;
	      case 'P': /* parity */
		i++;
		if(argv[i][0]=='N') parity=~PARENB;
		if(argv[i][0]=='E') parity=PARENB;
		if(argv[i][0]=='O') parity=PARODD;
		break;
	      case 'C': /* serial device */
		i++;
		if(argv[i]!=NULL) device=(unsigned char *)strdup(argv[i]);
		break;
	      case 'H':
		i++;
		hw_flow=1;
		break;
	      }
	  }
      }
    
    if( device == NULL )
      {
	fprintf(stderr,"You must specify a serial device using the -c " \
		"parameter.\n");
	exit(-1);
      }
}


/*----------------------------------------------------------------------------
 * FUNCTION: convert_baud
 * PARAMS  : String containing reuested baud rate
 * RETURNS : 0 on failure, baud rate constant on success
 * CALLS   : strcmp()
 * DESCR   : Convert a baud rate string to the equivalent numeric B*
 *           constant
 * --------------------------------------------------------------------------*/
unsigned int convert_baud(char *str)
{
  unsigned int i;

  if(!str)
    return 0; /* not found */

  for(i=0;i<BAUD_VALUE_COUNT;i++)
    {
      if(!strcmp(str,bstr[i]))
	return bvals[i];
    }

  return 0; /* not found */
}


/*----------------------------------------------------------------------------
 * FUNCTION: open_serial
 * PARAMS  : None
 * RETURNS : 0 on failure, 1 on success
 * CALLS   : open(), fcntl(), tcgetattr(), cfsetispeed(), cfsetospeed(),
 *           tcsetattr()
 * DESCR   : Open the serial port device with the paramters selected via
 *           config file and/or command line parameters.
 * --------------------------------------------------------------------------*/
int open_serial(void)
{
  struct termios options;
  
  serial_fd = open((const char *)device, O_RDWR | O_NOCTTY | O_NDELAY);
  if (serial_fd == -1)
    {
      return 0;
    }
  
  /* set non-blocking mode */
  fcntl(serial_fd, F_SETFL, FNDELAY);
    
  /* Get the current options for the port */
  tcgetattr(serial_fd, &options);
  
  /* Set the baud rate to BAUDRATE */
  if( cfsetispeed(&options, baudrate) ) return 0;
  if( cfsetospeed(&options, baudrate) ) return 0;
  
  /* Enable the receiver and set local mode  */
  options.c_cflag |= (CLOCAL | CREAD);
  
  switch( parity )
    {
    case ~PARENB: /* N1 */
      options.c_cflag &= ~PARENB;
      options.c_cflag &= ~CSTOPB;
      break;
    case PARENB: /* E1 */
      options.c_cflag |= PARENB;
      options.c_cflag &= ~PARODD;
      options.c_cflag &= ~CSTOPB;
      break;
    case PARODD: /* O1 */
      options.c_cflag |= PARENB;
      options.c_cflag |= PARODD;
      options.c_cflag &= ~CSTOPB;
      break;
    }
  
  options.c_cflag &= ~CSIZE;
  options.c_cflag |= CS8;
  
  /* Set no HW flow control */
  if(hw_flow)
    {
      options.c_cflag |= CRTSCTS;
    }
  else
    {
      options.c_cflag &= ~CRTSCTS;
    }
  
  /* Set no SW flow control, no CR/LF translation */
  options.c_iflag &= ~(IXON | IXOFF | IXANY | INLCR | ICRNL);
  
  /* Set raw input */
  options.c_lflag &= ~(ICANON | ECHO | ECHOE | ISIG);
  
  /* Set raw output */
  options.c_oflag &= ~OPOST;
  
  /* Set the new options for the port */
  if( tcsetattr(serial_fd, TCSAFLUSH, &options) ) return 0;

  return 1; /* set OK */
}


/*----------------------------------------------------------------------------
 * FUNCTION: open_console
 * PARAMS  : None
 * RETURNS : 1
 * CALLS   : tcgetattr(), tcsetattr()
 * DESCR   : Save a copy of the current terminal state. Set terminal to a
 *           a state that does no processing on the data coming in or
 *           going out.
 * --------------------------------------------------------------------------*/
int open_console(void)
{
  console_fd = STDIN_FILENO;

  /* get the terminal state - save a copy for exit */
  tcgetattr(console_fd,&ttysave);
  tcgetattr(console_fd,&ttystate);

  /* set terminal to 'raw' mode */
  ttystate.c_iflag &= ~(IGNBRK | BRKINT | PARMRK | ISTRIP
			  | INLCR | IGNCR | ICRNL | IXON);
  ttystate.c_oflag &= ~OPOST;
  ttystate.c_lflag &= ~(ECHO | ECHONL | ICANON | ISIG | IEXTEN);
  ttystate.c_cflag &= ~(CSIZE | PARENB);
  ttystate.c_cflag |= CS8;

  /* save terminal attributes */
  tcsetattr(console_fd, TCSANOW, &ttystate);

  return 1; /* set OK */
}

/*----------------------------------------------------------------------------
 * FUNCTION: terminal
 * PARAMS  : None
 * RETURNS : None
 * CALLS   : fprintf(), FD_ZERO(), FD_SET(), select(), read(), fflush(),
 *           toupper(), tcsetattr(), fgets(), open(), close(), perror(),
 *           strtok(), parse_hex2dec(), write(), system(), FD_ISSET(),
 *           memset(), fputc()
 * DESCR   : Main terminal loop:
 *               Look for available data on console
 *               If data, then
 *                  If COMMAND_KEY --> do command
 *               else
 *                  send data to serial device
 *               Look for available data on serial device
 *               If data, output to console in selected format
 * --------------------------------------------------------------------------*/
void terminal(void)
{
  struct timeval tv;
  fd_set fds;
  unsigned char data,prev_data,*ptr;
  int cmd_loop,i;
  unsigned char inbuf[INBUF_SIZE];
  unsigned char hex_buff[16];

  fprintf(stdout,"Terminal open. Press %s for command mode.\r\n",
	  COMMAND_KEY_STR);

  prev_data=0; /* used for 'history' to handle line endings */
  data=0;
  logging_mode=0;
  log_fd=-1;
  
  while(1)
    {
      /* Look for data on console */
      tv.tv_sec=0;
      tv.tv_usec=100;
      FD_ZERO(&fds);
      FD_SET(console_fd, &fds);  /* set bit for console */
      FD_SET(serial_fd, &fds);   /* set bit for serial device */
      select( serial_fd > console_fd ? (serial_fd+1) : (console_fd+1) ,
	      &fds, NULL, NULL, &tv );

      if( FD_ISSET(console_fd,&fds) ) /* we have console data */
	{
	  read(console_fd,&data,1);
	  if( data == COMMAND_KEY ) /* the command key was hit */
	    {
	      fprintf(stdout,"\r\nCmd: "); fflush(stdout);
	      cmd_loop=1;
	      while(cmd_loop)
		{
		  /* get command */
		  tv.tv_sec=0;
		  tv.tv_usec=100;
		  FD_ZERO(&fds);
		  FD_SET(console_fd, &fds);  /* set bit for console */
		  select( (console_fd+1) , &fds, NULL, NULL, &tv );
		  if( FD_ISSET(console_fd,&fds) )
		    {
		      read(console_fd,&data,1);

		      switch(toupper(data))
			{
			case COMMAND_KEY: /* exit command mode */
			  fprintf(stdout,"\r\n"); 
			  fflush(stdout);
			  cmd_loop=0;
			  break;
			case 'Q': /* quit terminal */
			  fprintf(stdout,"Quit\r\n");
			  return; 
			  break;
			case 'L': /* toggle log mode */
			  logging_mode = 1 - logging_mode;
			  if( logging_mode )
			    {
			      /* get filename from user */
			      fprintf(stdout,"Enable Log\r\n");
			      fprintf(stdout, "Enter file name. " \
				      "ENTER when done." \
				      "\r\n: ");
			      fflush(stdout);

			      /* go back into standard input mode */
			      tcsetattr(console_fd, TCSANOW, &ttysave);
			      fgets(inbuf,INBUF_SIZE-1,stdin);
			      for(i=0;i<strlen((const char *)inbuf);i++)
				{
				  if(inbuf[i] == '\r' ||
				     inbuf[i] == '\n')
				    inbuf[i]=0;
				}
			      /* return to console mode */
			      tcsetattr(console_fd, TCSANOW, &ttystate);

			      /* open log file for output */
			      if( (log_fd=open((const char *)inbuf,
					       O_RDWR|O_CREAT|O_APPEND,
					       S_IRUSR|S_IWUSR)) == -1 )
				{
				  fprintf(stdout,"Unable to open '%s' for "
					  "logging ",inbuf);
				  fflush(stdout);
				  perror("");
				  log_fd=-1;
				  logging_mode=0;
				}
			    }
			  else
			    {
			      if( log_fd != -1 )
				{
				  fprintf(stdout,"Disabling log\r\n");
				  close(log_fd);
				  log_fd=-1;
				}
			    }
			  cmd_loop=0;
			  break;
			case 'H': /* toggle hex printing mode */
			  hex_mode = 1 - hex_mode;
			  fprintf(stdout,"Hex Mode: %s\r\n",
				  hex_mode ? "Enabled" : "Disabled");
			  if(hex_mode)
			    {
			      fprintf(stdout,"\r\n");
			      hex_count=0;
			      memset(hex_buff,0,16);
			    }
			  fflush(stdout);
			  cmd_loop=0;
			  break;
			case 'R': /* allow entry of raw hex values to */
			          /* send out serial port */
			  fprintf(stdout,"Raw Hex\r\n");
			  fprintf(stdout, "Enter 2 digit hex values, " \
				  "separated by spaces. ENTER when done." \
				  "\r\n: ");
			  fflush(stdout);
			  
			  /* go back into standard input mode */
			  tcsetattr(console_fd, TCSANOW, &ttysave);
			  fgets(inbuf,INBUF_SIZE-1,stdin);
			  /* return to console mode */
			  tcsetattr(console_fd, TCSANOW, &ttystate);
			  
			  fprintf(stdout,"Sending: ");
			  fflush(stdout);
			  ptr=strtok(inbuf," \r\n");
			  while(ptr!=NULL)
			    {
			      data=parse_hex2dec(ptr);
			      fprintf(stdout,"%s ",ptr);
			      write(serial_fd,&data,1);
			      ptr=strtok(NULL," \r\n");
			    }
			  fprintf(stdout,"\r\n");
			  fflush(stdout);
			  cmd_loop=0;
			  break;
			case '?': /* show help */
			  fprintf(stdout,"Help\r\n\n\rAvailable Commands:" \
				  "\r\n\r\n");
			  fprintf(stdout,"CTL-Q  - Exit command mode\r\n");
			  fprintf(stdout,"Q      - Exit terminal\r\n");
			  fprintf(stdout,"H      - Toggle hex mode\r\n");
			  fprintf(stdout,"R      - Raw send hex values\r\n");
			  fprintf(stdout,"L      - Toggle logging\r\n");
			  fprintf(stdout,"D      - Display Macros\r\n");
			  fprintf(stdout,"1-9    - Execute macro\r\n");
			  fprintf(stdout,"\r\n");
			  fflush(stdout);
			  cmd_loop=0;
			  break;
			case 'D': /* display macro definitions */
			  fprintf(stdout,"Macro Definitions\r\n\r\n");
			  for(i=1;i<10;i++)
			    {
			      fprintf(stdout,"%1d: '%s'\r\n",i,
				      (macros[i] == NULL) ? "Not Defined" :
				      (char *)macros[i]);
			    }
			  fprintf(stdout,"\r\n");
			  cmd_loop=0;
			  break;
			case '1': /* catch requests to execute macros... */
			case '2':
			case '3':
			case '4':
			case '5':
			case '6':
			case '7':
			case '8':
			case '9':
			  fprintf(stdout,"\r--- EXECUTING MACRO %c ---\r\n",
				  data);
			  fprintf(stdout,"%s\r\n",macros[data-'0']);
			  /* restore console settings for external program */
			  tcsetattr(console_fd, TCSANOW, &ttysave);
			  close(serial_fd);
			  system((const char *)macros[data-'0']);
			  if( !open_serial() )
			    {
			      fprintf(stderr,"Error re-opening '%s' after "
				      "macro command.\n",device);
			      exit(1);
			    }
			  /* return to console mode after execution */
			  tcsetattr(console_fd, TCSANOW, &ttystate);
			  fprintf(stdout,"--- MACRO %c COMPLETE ---\r\n",
				  data);
			  cmd_loop=0;
			  break;
			}
		    }
		}
	    }
	  else /* not the COMMAND_KEY */
	    {
	      write(serial_fd,&data,1);
	      if( logging_mode )
		{
		  write(log_fd,&data,1);
		}
	    }
	}

      if( FD_ISSET(serial_fd,&fds) ) /* we have serial data */
	{
	  read(serial_fd,&data,1);
	  if( hex_mode )
	    {
	      if(hex_count>=16)
		{
		  fprintf(stdout,"\r\n");
		  hex_count=0;
		  memset(hex_buff,0,16);
		}
	      hex_buff[hex_count++]=data;
	      fprintf(stdout,"\r");
	      for(i=0;i<hex_count;i++)
		{
		  fprintf(stdout,"%02x ",hex_buff[i]);
		}
	      for(i=hex_count;i<16;i++)
		{
		  fprintf(stdout,"   ");
		}
	      fprintf(stdout,"    ");
	      for(i=0;i<hex_count;i++)
		{
		  fprintf(stdout,"%c",
			  (hex_buff[i]>31 & hex_buff[i]<128) ? 
			  hex_buff[i] : '.');
		}
	      fflush(stdout);
	      if( logging_mode )
		{
		  write(log_fd,&data,1);
		}
	    }
	  else if( line_end == END_CRLF )
	    { /* handle CRLF or LFCR line ends */
	      if( prev_data == '\n' && data != '\r' )
		{
		  fputc(data,stdout); fflush(stdout);
		  if( logging_mode )
		    {
		      write(log_fd,&data,1);
		    }
		}
	      else if( prev_data == '\r' && data != '\n' )
		{
		  fputc(data,stdout); fflush(stdout);
		  if( logging_mode )
		    {
		      write(log_fd,&data,1);
		    }
		}
	      else if( prev_data != '\r' && prev_data != '\n' )
		{
		  fputc(data,stdout); fflush(stdout);
		  if( logging_mode )
		    {
		      write(log_fd,&data,1);
		    }
		}
	    }
	  else
	    {
	      fputc(data,stdout); fflush(stdout);
	      if( logging_mode )
		{
		  write(log_fd,&data,1);
		}
	    }

	  prev_data=data; /* save byte for CRLF/LFCR handling */
	}
    }
}


/*----------------------------------------------------------------------------
 * FUNCTION: parse_hex2dec
 * PARAMS  : Pointer to string containing 2 hex digits
 * RETURNS : 0, or base-10 value of 2 hex digits
 * CALLS   : None
 * DESCR   : Convert the hexidecimial 2 digit number pointed to by 'ptr' into
 *           a base-10 value. 'ptr' must be pointing to most significant digit
 * --------------------------------------------------------------------------*/
unsigned char parse_hex2dec(char *ptr)
{
  unsigned char temp;

  if( ptr == NULL )
    return 0;

  temp=0;

  if( (*ptr >= '0') && (*ptr <= '9') )
    temp = (*ptr - '0') * 16;
  if( (*ptr >= 'A') && (*ptr <= 'F') )
    temp = (*ptr - 55) * 16;
  if( (*ptr >= 'a') && (*ptr <= 'f') )
    temp = (*ptr - 87) * 16;

  ptr++;

  if( (*ptr >= '0') && (*ptr <= '9') )
    temp += (*ptr - '0');
  if( (*ptr >= 'A') && (*ptr <= 'F') )
    temp += (*ptr - 55);
  if( (*ptr >= 'a') && (*ptr <= 'f') )
    temp += (*ptr - 87);

  return temp;
}


/*----------------------------------------------------------------------------
 * FUNCTION: set_defaults
 * PARAMS  : None
 * RETURNS : None
 * CALLS   : None
 * DESCR   : Set default values for serial device parameters. Initialize
 *           macro strings to no value
 * --------------------------------------------------------------------------*/
void set_defaults(void)
{
  unsigned int i;

  /* set defaults */
  baudrate = B9600;
  device = NULL;
  databits = CS8;
  parity = ~PARENB;
  stopbits = ~CSTOPB;
  line_end=END_NL;
  hex_mode=0;
  hw_flow=0;

  for(i=0;i<10;i++) macros[i]=NULL;
}


/*----------------------------------------------------------------------------
 * FUNCTION: dot_file
 * PARAMS  : None
 * RETURNS : None
 * CALLS   : getcwd(), strncat(), access(), read_dot(), getenv(),
 *           getpwuid(), strncpy()
 * DESCR   : Look for MACRO_FILE in the current working directory. If not
 *           there, look for it in the users $HOME directory. When found,
 *           call dot_read() to read the contents.
 * --------------------------------------------------------------------------*/
void dot_file(void)
{
  char cwd[_POSIX_PATH_MAX], *homedir;
  struct passwd *pwent;

  if (getcwd(cwd, sizeof(cwd)) != NULL)
    {
      strncat(cwd,"/",sizeof(cwd)-strlen(cwd));
      strncat(cwd,MACRO_FILE,sizeof(cwd)-strlen(cwd));
      if( access(cwd,R_OK) != -1 ) /* MACRO_FILE exists in CWD.. use it */
	{
	  read_dot(cwd);
	  return;
	}
    }

  /* at this point, cant get CWD, or local MACRO_FILE doesn't exist. */
  /* look in home directory. */
  homedir=NULL;
  if ((homedir = getenv("HOME")) == NULL) 
    {
      pwent = (struct passwd *)getpwuid(getuid());
      if( pwent != NULL )
	homedir = pwent->pw_dir;
    }
  if( homedir != NULL )
    {
      strncpy(cwd,homedir,sizeof(cwd));
      strncat(cwd,"/",sizeof(cwd)-strlen(cwd));
      strncat(cwd,MACRO_FILE,sizeof(cwd)-strlen(cwd));
      if( access(cwd,R_OK) != -1 ) /* MACRO_FILE exists in HOME.. use it */
	{
	  read_dot(cwd);
	  return;
	}
    }

}


/*----------------------------------------------------------------------------
 * FUNCTION: read_dot
 * PARAMS  : Path to dot file to read
 * RETURNS : None
 * CALLS   : fopen(), fprintf(), fgets(), strtok(), strncasecmp(),
 *           convert_baud(), strncpy(), fclose()
 * DESCR   : Read configuration / macro file. Blank lines, and lines
 *           prefixed with '#' are ignored. Baud rate may be set using
 *           'baudrate=<numeric value>'; serial device may be set using
 *           'device=<device string>'; macros 1-9 may be set using 'N=string'
 * --------------------------------------------------------------------------*/
void read_dot(char *file)
{
  FILE *dotfile;
  char parsebuf[_POSIX_PATH_MAX], *ptr;
  unsigned int i,macro_num;
  
  if(file==NULL)
    return;

  if((dotfile=fopen(file,"rt")) == NULL )
    {
      fprintf(stderr,"%s exists, but cannot be read!\n",file);
    }
  while(fgets(parsebuf,_POSIX_PATH_MAX,dotfile)!=NULL)
    {
      for(i=0;i<strlen(parsebuf);i++)
	if(parsebuf[i]=='\n' || parsebuf[i]=='\r')
	  parsebuf[i]=0;

      if( (parsebuf[i]!='#') && (strlen(parsebuf)>0) )
	{
	  /* it's not a comment or a blank line */
	  if((ptr=strtok(parsebuf,"=")) != NULL)
	    {
	      if(strncasecmp(ptr,"baudrate",8)==0)
		{
		  if((ptr=strtok(NULL,"="))!=NULL)
		    {
		      if(!(baudrate=convert_baud(ptr)))
			{
			  fprintf(stderr,
				  "%s: Unable to parse baudrate '%s'\n",
				  MACRO_FILE,
				  ptr);
			}
		      else
			{
			  fprintf(stderr,"%s: Baud rate = '%s'\n",MACRO_FILE,
				  ptr);
			}
		    }
		}
	      if(strncasecmp(ptr,"device",6)==0)
		{
		  if((ptr=strtok(NULL,"="))!=NULL)
		    {
		      device=strdup(ptr);
		      fprintf(stderr,"%s: Device = '%s'\n",MACRO_FILE,device);
		    }		  
		}
	      if( (ptr[0]>='1')&&(ptr[0]<='9') )
		{ 
		  /* macro definition */
		  macro_num = ptr[0] - '0';
		  if((ptr=strtok(NULL,"="))!=NULL)
		    {
		      macros[macro_num]=malloc(sizeof(char)*_POSIX_PATH_MAX);
		      strncpy(macros[macro_num],ptr,_POSIX_PATH_MAX);
		    }
		}
	    }
	} /* end its not a comment... */
    }
  fclose(dotfile);
}


/*----------------------------------------------------------------------------
 * FUNCTION: substitute_in_macros
 * PARAMS  : None
 * RETURNS : None
 * CALLS   : strstr(), strlen(), memmove(), memcopy()
 * DESCR   : Make substitutions in macro strings for buadrate ('$$B'), 
 *           serial device name ('$$D')
 * --------------------------------------------------------------------------*/
void substitute_in_macros(void)
{
  unsigned char i,j;
  char buf[_POSIX_PATH_MAX],*p,*bptr;

  bptr="";
  for(i=0;i<BAUD_VALUE_COUNT;i++)
    {
      if(bvals[i] == baudrate)
	bptr=bstr[i];
    }
  
  for(i=1;i<10;i++)
    {
      if(macros[i]!=NULL)
	{
	  /* substitute device name */
	  while( (p=strstr(macros[i],"$$D")) && 
		 ((strlen(macros[i])+strlen(device))<_POSIX_PATH_MAX) )
	    {
	      /* shift string down */
	      memmove(p+strlen(device), /* dest */
		      p+3, /* src */
		      (strlen(macros[i])-(p-macros[i])-3) );
	      /* copy device into place */
	      memcpy(p,device,strlen(device));
	    }

	  /* substitute baud rate */
	  while( (p=strstr(macros[i],"$$B")) && 
		 ((strlen(macros[i])+strlen(bptr))<_POSIX_PATH_MAX) )
	    {
	      /* shift string down */
	      memmove(p+strlen(bptr), /* dest */
		      p+3, /* src */
		      (strlen(macros[i])-(p-macros[i])-3) );
	      /* copy device into place */
	      memcpy(p,bptr,strlen(bptr));
	    }
	}
    }
}
